from fragmenstein import Laboratory
from rdkit import Chem
from rdkit.Chem import AllChem
import contextlib
import pandas as pd
import warnings
from .query import QueryArthor
from typing import Dict, Union, Optional


def prep_for_laboratory(query_mol: Chem.Mol,
                        template_mol: Chem.Mol,
                        experiment_name:str='') \
        -> pd.DataFrame:
    """
    This function is a wrapper for ``create_laboratory_df`` and ``assess_experiment``.
    It creates a dataframe of analogues and assesses the experiment.
    The dataframe can be saved and run in a different node.
    The logic being that it is often required to tweak the SMARTS query to get a good amount of analogues,
    whereas the placement can happen in the background.
    To run one to check run ``assess_experiment``.

    :param query_mol:
    :param template_mol:
    :param experiment_name:
    :return:
    """

    analogs = create_laboratory_df(query_mol, template_mol, experiment_name)
    print('No analogues found' if len(analogs) == 0 else f'{len(analogs)} analogues found')
    return analogs
def run_experiment(query: Chem.Mol, template: Chem.Mol, pdb_block, experiment_name: str=''):
    """
    Calls ``create_laboratory_df`` and ``Laboratory.place``.

    :param query:
    :param template:
    :param pdb_block:
    :param experiment_name:
    :return:
    """
    placements = pd.DataFrame()
    with contextlib.suppress(KeyError):
        tobeplaced = create_laboratory_df(query, template, experiment_name=experiment_name)
        placements = Laboratory(pdb_block).place(tobeplaced)
    placements['experiment'] = experiment_name
    return placements

def create_laboratory_df(query_mol: Chem.Mol, template_mol: Chem.Mol, experiment_name: str= '') -> pd.DataFrame:
    """
    Given a query molecule and a template molecule, return a dataframe of analogues
    by searching Arthor.
    This dataframe is ready to be passed to ``Laboratory.place``.
    It has the columns:

    * name: the name of the analogue
    * hits: a list of the template molecule
    * custom_map: a dictionary of the custom map from the template to the analogue
    * experiment: the name of the experiment

    :param query_mol:
    :param template_mol:
    :param experiment_name:
    :return:
    """
    # ## Sanitize
    if not experiment_name and query_mol.HasProp('experiment'):
        experiment_name = query_mol.GetProp('experiment')
    query_mol.SetProp('experiment', experiment_name)
    query_mol.SetProp('template', template_mol.GetProp('_Name'))
    # ## Run search
    arthor = QueryArthor()
    df = arthor.retrieve(Chem.MolToSmarts(query_mol), ['real-database-22q1'])
    query_mol.UpdatePropertyCache()
    #AllChem.SanitizeMol(query)
    # exit early if no hits
    if len(df) == 0:
        return pd.DataFrame()
    try:
        df = df.rename(columns={'id': 'name'})
        df['hits'] = df.smiles.apply(lambda s: [template_mol])
        mapper = lambda smiles: {template_mol.GetProp('_Name'): get_custom_map(query_mol, template_mol, smiles)}
        df['custom_map'] = df.smiles.apply(mapper)
        df['experiment'] = experiment_name
        return df
    except Exception as error:
        print(error.__class__.__name__, error)
        return df

def assess_experiment(query: Chem.Mol, template: Chem.Mol, pdb_block: str) -> Chem.Mol:
    """
    Given a query molecule and a template molecule, assess the experiment by placing the query molecule

    :param query:
    :param template:
    :param pdb_block:
    :return:
    """
    name = query.GetProp('experiment') if query.HasProp('experiment') else 'test'
    query.UpdatePropertyCache()
    AllChem.SanitizeMol(query, catchErrors=True)
    assert query.GetConformer()
    cm = get_custom_map(query, template, Chem.MolToSmiles(template))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        minimised = Laboratory.Victor([template], pdb_block=pdb_block).place(Chem.MolToSmiles(template),
                                                                        long_name=name,
                                                                        custom_map=cm).minimized_mol
        assert minimised, 'Nothing returned'
        return minimised



def get_custom_map(shared_query: Chem.Mol, template: Chem.Mol, final: Union[str, Chem.Mol]) -> Dict[int, int]:
    """
    Given a template molecule and a final molecule and a shared query molecule,
    return the custom map going from atom indices in the template molecule to the final molecule in the smiles string.

    :param shared_query:
    :param template:
    :param final:
    :return:
    """
    query2template = dict(enumerate(template.GetSubstructMatch(shared_query)))
    if isinstance(final, str):
        final = Chem.MolFromSmiles(final)
    query2analog_iter = enumerate(final.GetSubstructMatch(shared_query))
    return {query2template[qi]: ai for qi, ai in query2analog_iter}



__all__ = ['SmartMonsterHandler']

from rdkit import Chem
from fragmenstein import Monster
from contextlib import suppress
from rdkit.rdBase import BlockLogs


class SmartMonsterHandler:
    """
    Given hits and a SMARTS query w/ dot separated patterns,
    wherein each sequentially matches the hits

    ... code-block:: python
        sm = SmartMonsterHandler([hit1, hit2], query, joining_cutoff=10)
        sm(matches.mol[0])

    or

    ... code-block:: python
        sm = SmartMonsterHandler([hit1, hit2], query, joining_cutoff=10)
        results = matches.mol.apply(sm)
        matches['success'] = results.loc[~results.isna()].apply(operator.itemgetter('success'))
        matches['ddG'] = results.loc[~results.isna()].apply(operator.itemgetter('ddG'))
        matches['minimized_mol'] = results.loc[~results.isna()].apply(operator.itemgetter('mol'))

"""

    def __init__(self, hits, query, **monster_options):
        self.hits = hits
        self.monster_options = monster_options
        self.names = [h.GetProp('_Name') for h in self.hits]
        self.smols = [Chem.MolFromSmarts(q) for q in query.split('.')]
        self.mappings = [h.GetSubstructMatch(smol) for h, smol in zip(self.hits, self.smols)]
        assert all([len(m) > 0 for m in self.mappings]), 'The query does not match the hits'
        self._exception = Exception

    def map(self, mol):
        return {n: dict(zip(m, mol.GetSubstructMatch(s))) for n, m, s in zip(self.names, self.mappings, self.smols)}

    def __call__(self, mol):
        monster = Monster(self.hits, **self.monster_options)
        mapping = self.map(mol)
        with suppress(self._exception), BlockLogs():
            monster.place(mol, custom_map=mapping)
            used_map = monster.convert_origins_to_custom_map()
            success = monster.mmff_minimize(allow_lax=False)
            ddG = monster.MMFF_score()
            if any([len(set(mapping[name].items()) - set(used_map[name].items())) != 0 for name in self.names]):
                success = False
                ddG = float('nan')
            return {'mol': monster.positioned_mol,
                    'custom_map': mapping,
                    'success': success,
                    'ddG': ddG
                    }
        return {'mol': mol, 'custom_map': mapping, 'success': False, 'ddG': float('nan')}


from .query import QueryArthor

import contextlib
import pandas as pd
import warnings
from rdkit import RDLogger, Chem
from rdkit.Chem import AllChem, Draw
from IPython.display import display
from typing import Dict, Union


def silence_rdkit():
    """
    Silence RDKit warnings
    """
    logger = RDLogger.logger()
    logger.setLevel(RDLogger.CRITICAL)

def show_experiment(query3d: Chem.Mol, experiment_name: str='', save: bool=False):
    """
    Show the 2D structure of the query molecule, experiment_name and its SMARTS pattern.

    :param query3d:
    :param experiment_name:
    :return:
    """
    query = Chem.Mol(query3d)
    if not experiment_name:
        experiment_name = query.GetProp('experiment') if query.HasProp('experiment') else 'test'
    AllChem.Compute2DCoords(query)
    print(Chem.MolToSmarts(query))
    display(query)
    if save:
        Draw.MolToFile(query, f'{experiment_name}.png')
    return None




__all__ = ['enquire', 'get_charge_string']

from rdkit import Chem
from typing import Mapping, Sequence, Union, List
from IPython.display import Image as IPyImage
from PIL import Image as PILImage
import requests
import io
import warnings


def enquire(mol: Chem.Mol,
                replacements: Mapping[int, Union[str, None]] = {},
                rgroups: Sequence[int] = {},
                generic_arocarbons: bool = False) -> Chem.Mol:
    """
    Given a molecule, convert it to a molecule with query atoms,
    with correct element, number of hydrogens and charge,
    but with overridden values as given by ``replacements`` argument,
    which accepts a dictionary of index (int) to SMARTS (str) or None (delete atom).

    A ``Chem.QueryAtom`` is a special atom with encoded ambiguity.
    ``Chem.MolFromSmarts`` will create a molecule with query atoms, but not a regular mol.
    A query atom has the additional methods
    ``.HasQuery()``, ``.GetQueryType()``, ``.DescribeQuery()``.
    cannot be instantiated from Python, but can be using ``Chem.AtomFromSmarts``.

    Additionally, any atom idx in argument ``rgroups`` will get a sequential isotope number from one
    and property 'R-group' of R + off-by-one number.
    If this index was not in replacements, then the SMARTS will have one connection more than there are
    and one implict hydrogen less.

    This function requires ``mol`` to have implicit hydrogens.

    ..code-block::python
       queried:Chem.Mol = enquire(Chem.MolFromSmiles('c1cnccc1'), {2: '[c,n]'})
       Chem.MolToSmarts(queried)
       # '[c&H1]1:[c&H1]:[c,n]:[c&H1]:[c&H1]:[c&H1]:1'

    Note 1. ``atom.GetSmarts()`` is a method, but it could return '[N+]' already,
    which is complicated to deal with as appending at given positions may muck things up.
    And certain SMILES are not legal, for example 'CC[O+H2]' should be 'CC[OH2+]'

    This method was formerly called ``querimonate``.
    Querimony is a complaint. There is no verb form. This is a jocular neologism,
    as RDKit will complain... But the name is too confusing to write.
    """
    removals: List[int] = []
    mod = Chem.RWMol(mol)
    atom: Chem.Atom
    for smarts in replacements.values():
        if not smarts:
            continue
        else:
            assert Chem.MolFromSmarts(smarts) is not None, f'Could not parse SMARTS {smarts}'
    for atom in mod.GetAtoms():
        assert atom.GetNumRadicalElectrons() == 0, 'This molecule has a radical'
        idx: int = atom.GetIdx()
        n_Hs: int = atom.GetImplicitValence() + atom.GetNumExplicitHs()
        symbol: str = atom.GetSymbol().lower() if atom.GetIsAromatic() else atom.GetSymbol()
        scharge: str = get_charge_string(atom)
        # pick relevant SMARTS
        if idx in replacements:
            # user override replacement
            if isinstance(replacements[idx], str) and replacements[idx] != '':
                smarts: str = replacements[idx]
            else:
                removals.append(idx)
                smarts: str = ''
        elif idx in rgroups:
            # user override rgroups
            assert n_Hs == 0, 'R-group requested for zero Hs. Use charge or change element via ``replacement``'
            n_Xs: int = len(atom.GetNeighbors())
            smarts = f'[{symbol}H{n_Hs - 1}X{n_Xs + 1}]'
        elif not generic_arocarbons or symbol != 'c' or scharge != '':
            # keep as is
            smarts = f'[{symbol}H{n_Hs}{scharge}]'
        elif n_Hs == 1:
            # degenerate aromatic carbon w/ one hydrogen
            smarts = '[aH0X2,aH1X3]'
        else:
            # degenerate aromatic carbon w/ zero hydrogens
            smarts = '[aH0]'
        # swap
        if smarts == '':
            pass  # delete
        elif isinstance(atom, Chem.QueryAtom):
            # weird it is already a query atom
            atom.SetQuery(smarts)
        else:
            mod.ReplaceAtom(idx, Chem.AtomFromSmarts(smarts), preserveProps=True)
    for r, idx in enumerate(rgroups):
        atom = mod.GetAtomWithIdx(idx)
        atom.SetIsotope(r + 1)
        atom.SetProp('R-group', f'R{r + 1}')
    mod.BeginBatchEdit()
    for idx in removals:
        mod.RemoveAtom(idx)
    mod.CommitBatchEdit()
    return mod.GetMol()

def querimonate(*args, **kwargs):
    """
    See ``enquire``.
    """
    warnings.warn('`querimonate` has been renamed `eniquire`. Use `enquire` instead.', DeprecationWarning)
    return enquire(*args, **kwargs)

def get_charge_string(atom: Chem.Atom) -> str:
    """
    Returns a plus or minus string for the charge of an atom.
    :param atom: Chem.Atom
    :return:
    """
    # charge
    charge: int = atom.GetFormalCharge()
    if charge > 0:
        return '+' * abs(charge)
    elif charge < 0:
        return '-' * abs(charge)
    else:
        return ''


def retrieve_smartsplus(smarts: Union[str, Chem.Mol], PIL_image=True, **options) -> Union[IPyImage, PILImage.Image]:
    """
    Given a SMARTS query, retrieve the image from https://smarts.plus.
    The returned object is an IPython.display.Image not a PIL.Image.
    If using this image remember to cite it as
    "SMARTSviewer smartsview.zbh.uni-hamburg.de, ZBH Center for Bioinformatics, University of Hamburg"

    :param smarts: SMARTS query or Chem.Mol
    :param PIL_image: return PIL.Image instead of IPython.display.Image
    :param options: See https://smarts.plus/rest
    :return:
    """
    if isinstance(smarts, Chem.Mol):
        q = smarts
        smarts: str = Chem.MolFromSmarts(q)
    # retrieve from smarts.plus
    response: requests.Response = requests.get('https://smarts.plus/smartsview/download_rest', # noqa to sensitive data in options
                                               {'smarts': smarts, **options}
                                              )
    png_binary: bytes = response.content
    if PIL_image:
        return PILImage.open(io.BytesIO(png_binary))
    else:
        return IPyImage(data=png_binary)



__all__ = ['QueryArthor']

import requests
import warnings
import pandas as pd
from typing import List
from rdkit import Chem
from rdkit.Chem import PandasTools, Draw, AllChem


class QueryArthor:
    """
    Query class for Arthorian Quest.
    It queries arthor.docking.org via ``.retrieve``

    See https://arthor.docking.org/api.html for the API endpoints used

    .. code-block:: python
        Query().retrieve('[CH3]-[CH2X4]', ['BB-50-22Q1'])
    """

    enamine_dbs = ['BB-ForSale-22Q1', 'MADE-BB-23Q1-770M', 'REAL-Database-22Q1']

    def __init__(self, base_url: str = 'https://arthor.docking.org/'):
        self.base_url = base_url

    @property
    def dbs(self):
        return pd.DataFrame(requests.get(self.base_url + 'dt/data').json())

    def retrieve(self, query: str, dbnames: List[str], search_type='SMARTS', length=1_000):
        """
        Returns a dataframe of the results of the query,
        with fields:

        * N_RB: number of rotatable bonds
        * N_HA: number of heavy atoms

        :param query: SMARTS query
        :param dbnames: list of names (see self.dbs)
        :return:
        """
        dbname: str = ','.join(dbnames)
        if isinstance(query, Chem.Mol):
            query = Chem.MolToSmarts(query)
        response: requests.Response = requests.get(self.base_url + f'/dt/{dbname}/search',
                                                   dict(query=query,
                                                        type=search_type,
                                                        length=length)
                                                   )
        data: dict = response.json()
        if data.get("message", '') == "SMARTS query is always false!":
            warnings.warn(f"SMARTS query {query} is always false")
            return pd.DataFrame()
        if data.get('warning',''):
            warnings.warn(data['warning'])
        if not data.get('recordsTotal', False):
            warnings.warn(f"SMARTS query {query} returned no matches")
            return pd.DataFrame()
        matches = pd.DataFrame(data['data'], columns=['idx', 'smiles_id', 'empty', 'something', 'db'])
        if len(matches) == 0:  # empty
            return matches
        matches['id'] = matches.smiles_id.str.split(expand=True)[1]
        matches['smiles'] = matches.smiles_id.str.split(expand=True)[0]
        matches.drop_duplicates('id')
        PandasTools.AddMoleculeColumnToFrame(matches, 'smiles', 'mol', includeFingerprints=True)
        matches = matches.loc[~matches.mol.isnull()]
        # tabs?
        matches['db'] = matches.db.str.strip()
        matches['N_RB'] = matches.mol.apply(AllChem.CalcNumRotatableBonds)
        matches['N_HA'] = matches.mol.apply(AllChem.CalcNumHeavyAtoms)
        return matches.sort_values('N_HA').reset_index(drop=True)



from setuptools import setup, find_packages
from warnings import warn
from importlib import util
import os

this_directory = os.path.abspath(os.path.dirname(__file__))

if os.path.exists(os.path.join(this_directory, 'README.md')):
    with open(os.path.join(this_directory, 'README.md'), 'r') as f:
        long_description = f.read()
else:
    long_description = '''## Arthorian quest\n'''

if os.path.exists(os.path.join(this_directory, 'requirements.txt')):
    with open(os.path.join(this_directory, 'requirements.txt'), 'r') as f:
        requirements = [line.split('#')[0].strip() for line in f.readlines()]
        requirements = [line for line in requirements if line]
else:
    requirements = []

setup(
    name='arthorian-quest',
    version='0.2',
    description='using Arthor and filtering the results with Fragmenstein',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    #extras_require={'jupyter': ['jupyter']},
    url='https://github.com/matteoferla/arthorian-quest',
    license='MIT',
    author='Matteo Ferla',
    author_email='matteo.ferla@gmail.com',
    classifiers=[ # https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha', # Development Status :: 5 - Production/Stable
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)